package ap.parameters;

import ap.util.Debug

object Settings {
  
  private val AC = Debug.AC_PARAMETERS
  
  /**
   * An boolean option of the kind "+opt" or "-opt"
   */
  private object Opt {
    def unapply(arg : String) : Option[(String, Boolean)] =
      if (arg startsWith "+")
        Some(arg substring 1, true)
      else if (arg startsWith "-")
        Some((arg substring 1, false))
      else
        None
  }
  
  /**
   * An assignment of the kind "-var13=4" or "-xxx=hallo". The assigned value
   * is here only decoded to a string
   */  
  private object ValueOpt {
    def unapply(arg : String) : Option[(String, String)] =
      if ((arg startsWith "-") && (arg contains "=")) {
        val i = arg indexOf "="
        Some(arg.substring(1, i), arg substring (i+1))
      } else {
        None
      }
  }

  /**
   * A value that represents an integer, like in "-var13=4"
   */  
  private object IntVal {
    def unapply(arg : String) : Option[Int] =
      try {
       	Some(arg.toInt) 
      } catch {
        case _ : NumberFormatException => None
      }
  }
    
  def fromArguments(args : Array[String]) : (Settings, Seq[String]) = {
    val settings = new Settings
    val inputs = new scala.collection.mutable.ArrayBuffer[String]
    
    for (arg <- args) arg match {
      case Opt("printTree", value) =>
        Param.PRINT_TREE.set(settings, value)
      case Opt("assert", value) =>
        Param.ASSERTIONS.set(settings, value)
      case Opt("simplifyConstraints", value) =>
        Param.SIMPLIFY_CONSTRAINTS.set(settings, value)
      case Opt("traceConstraintSimplifier", value) =>
        Param.TRACE_CONSTRAINT_SIMPLIFIER.set(settings, value)
      case ValueOpt("timeout", IntVal(value)) =>
        Param.TIMEOUT.set(settings, value)
      case Opt(_, _) =>
        throw new UnknownArgumentException(arg)
      case _ => inputs += arg
    }
    
    (settings, inputs)
  }
  
  class UnknownArgumentException(arg : String)
        extends Exception("Unknown argument: " + arg)
  
}

class Settings {

  private val paramMap = new scala.collection.mutable.HashMap[Param, Any]
  
  protected[parameters] def apply(p : Param) : Any = (paramMap get p) match {
    case Some(value : Any) => {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(Settings.AC, value.isInstanceOf[p.Value])
      //////////////////////////////////////////////////////////////////////////
      value
    }
    case None => p.defau
  }
  
  protected[parameters] def +=(paramPair : (Param, Any)) : Settings = {
    val (p, value) = paramPair
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(Settings.AC, value.isInstanceOf[p.Value])
    ////////////////////////////////////////////////////////////////////////////
    paramMap += paramPair
    this
  }
  
}
