package ap.terfor;

import scala.collection.mutable.ArrayBuffer

object Formula {
  
  /**
   * Check whether the given sequence of formulas contains unsatisfiable
   * elements (in this case, a singleton sequence containing only this element
   * is returned), and otherwise remove all tautologies from it.
   */
  def filterForConj[A <: Formula](fors : Iterator[A]) : RandomAccessSeq[A] = {
    val nonTrivialFors = new ArrayBuffer[A]
    for (c <- fors) {
      if (c.isFalse) return Array(c)
      if (!c.isTrue) nonTrivialFors += c
    }
    
    nonTrivialFors
  }
  
  def conj[A <: Formula](fors : Iterator[A],
                         trueFor : A, comb : (RandomAccessSeq[A]) => A) : A = {
    val nonTrivialFors = filterForConj(fors)
    
    nonTrivialFors.size match {
      case 0 => trueFor
      case 1 => nonTrivialFors(0)
      case _ => comb(nonTrivialFors)
    }
  }
}

abstract class Formula extends TerFor {
  
  /** Return <code>true</code> if this formula is obviously always true */
  def isTrue : Boolean
  
  /** Return <code>true</code> if this formula is obviously always false */
  def isFalse : Boolean
  
}
