package ap.terfor.substitutions;

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.util.Debug

object ComposeSubsts {
  
  private val AC = Debug.AC_SUBSTITUTIONS

  def apply(substs : Iterable[Substitution], order : TermOrder) : Substitution =
    apply(substs.elements, order)

  def apply(substs : Iterator[Substitution], order : TermOrder) : Substitution =
    (new IdentitySubst (order).asInstanceOf[Substitution] /: substs)(
      (combination, subst) => apply(combination, subst, order))
  
  def apply(secondSubst : Substitution,
            firstSubst : Substitution,
            order : TermOrder) : Substitution = {
    ////////////////////////////////////////////////////////////////////////////
    // a sufficient, but not a necessary condition for getting the composed
    // substitution work correctly. This could be weakened
    Debug.assertPre(AC,
                    firstSubst.order == order && secondSubst.order == order)
    ////////////////////////////////////////////////////////////////////////////

    (secondSubst, firstSubst) match {
    case (_, _ : IdentitySubst) => secondSubst
    case (_ : IdentitySubst, _) => firstSubst
    case _ => new ComposeSubsts(secondSubst, firstSubst, order)
    }
  }
  
}

/**
 * Function composition for two substitutions
 */
class ComposeSubsts private (secondSubst : Substitution,
                             firstSubst : Substitution,
                             protected [substitutions] val order : TermOrder)
      extends Substitution {

  //////////////////////////////////////////////////////////////////////////////
  // a sufficient, but not a necessary condition for getting the composed
  // substitution work correctly. This could be weakened
  Debug.assertCtor(ComposeSubsts.AC,
                   firstSubst.order == order && secondSubst.order == order)
  //////////////////////////////////////////////////////////////////////////////     

  protected[substitutions] def passQuantifiers(num : Int) : Substitution =
    new ComposeSubsts (secondSubst.passQuantifiers(num),
                       firstSubst.passQuantifiers(num),
                       order)
   
  /**
   * For composition, we simply first apply the first substitution and then
   * the second
   */
  def apply(t : Term) : Term =
    secondSubst(firstSubst(t))
     
  def apply(lc : LinearCombination) : LinearCombination =
    secondSubst(firstSubst(lc))

  def apply(conj : EquationConj) : EquationConj =
    secondSubst(firstSubst(conj))
                                         
  def apply(negConj : NegEquationConj) : NegEquationConj =
    secondSubst(firstSubst(negConj))

  def apply(conj : InEqConj) : InEqConj =
    secondSubst(firstSubst(conj))

  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean =
    firstSubst.isIdentityOn(t) && secondSubst.isIdentityOn(t)

  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : ComposeSubsts =
    new ComposeSubsts(secondSubst.sortBy(newOrder),
                      firstSubst.sortBy(newOrder),
                      newOrder)
 
  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = "" + secondSubst + " . " + firstSubst

}
